@@ -1,4 +1,10 @@
+# copyright ############################### #
+# This file is part of the Xtrack Package.  #
+# Copyright (c) CERN, 2021.                 #
+# ######################################### #
+
 from pathlib import Path
+from tkinter import N
 
 import numpy as np
 from scipy.special import factorial
@@ -8,6 +14,7 @@ import xpart as xp
 
 from ..base_element import BeamElement
 from ..general import _pkg_root
+from ..interal_record import RecordIndex, RecordIdentifier
 
 class ReferenceEnergyIncrease(BeamElement):
 
@@ -19,13 +26,13 @@ class ReferenceEnergyIncrease(BeamElement):
     _xofields = {
         'Delta_p0c': xo.Float64}
 
+    _extra_c_sources = [
+        _pkg_root.joinpath('beam_elements/elements_src/referenceenergyincrease.h')]
+
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(Delta_p0c=-self.Delta_p0c,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-ReferenceEnergyIncrease.XoStruct.extra_sources = [
-        _pkg_root.joinpath('beam_elements/elements_src/referenceenergyincrease.h')]
-
 
 class Drift(BeamElement):
     '''Beam element modeling a drift section. Parameters:
@@ -36,13 +43,15 @@ class Drift(BeamElement):
     _xofields = {
         'length': xo.Float64}
     isthick=True
+    behaves_like_drift=True
+
+    _extra_c_sources = [_pkg_root.joinpath('beam_elements/elements_src/drift.h')]
 
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(length=-self.length,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-Drift.XoStruct.extra_sources = [
-        _pkg_root.joinpath('beam_elements/elements_src/drift.h')]
+
 
 class Cavity(BeamElement):
     '''Beam element modeling an RF cavity. Parameters:
@@ -58,6 +67,10 @@ class Cavity(BeamElement):
         'lag': xo.Float64,
         }
 
+    _extra_c_sources = [
+        _pkg_root.joinpath('headers/constants.h'),
+        _pkg_root.joinpath('beam_elements/elements_src/cavity.h')]
+
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(
                               voltage=-self.voltage,
@@ -65,10 +78,6 @@ class Cavity(BeamElement):
                               lag=self.lag,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-Cavity.XoStruct.extra_sources = [
-        _pkg_root.joinpath('headers/constants.h'),
-        _pkg_root.joinpath('beam_elements/elements_src/cavity.h')]
-
 
 class XYShift(BeamElement):
     '''Beam element modeling an transverse shift of the reference system. Parameters:
@@ -82,16 +91,14 @@ class XYShift(BeamElement):
         'dy': xo.Float64,
         }
 
+    _extra_c_sources = [
+        _pkg_root.joinpath('beam_elements/elements_src/xyshift.h')]
+
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(
                               dx=-self.dx, dy=-self.dy,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-XYShift.XoStruct.extra_sources = [
-        _pkg_root.joinpath('beam_elements/elements_src/xyshift.h')]
-
-
-## ELECTRON LENS
 
 class Elens(BeamElement):
 # if array is needed we do it like this
@@ -106,22 +113,29 @@ class Elens(BeamElement):
                'residual_kick_y': xo.Float64
               }
 
-    def __init__(self,  inner_radius  = None,
-                        outer_radius  = None,
-                        current       = None,
-                        elens_length  = None,
-                        voltage       = None,
+    _extra_c_sources = [
+        _pkg_root.joinpath('beam_elements/elements_src/elens.h')]
+
+    def __init__(self,  inner_radius = 1.,
+                        outer_radius = 1.,
+                        current      = 0.,
+                        elens_length = 0.,
+                        voltage      = 0.,
                         residual_kick_x = 0,
                         residual_kick_y = 0,
+                        _xobject = None,
                         **kwargs):
-        super().__init__(**kwargs)
-        self.inner_radius    = inner_radius
-        self.outer_radius    = outer_radius
-        self.current         = current
-        self.elens_length    = elens_length
-        self.voltage         = voltage
-        self.residual_kick_x   = residual_kick_x
-        self.residual_kick_y   = residual_kick_y
+        if _xobject is not None:
+            super().__init__(_xobject=_xobject)
+        else:
+            super().__init__(**kwargs)
+            self.inner_radius    = inner_radius
+            self.outer_radius    = outer_radius
+            self.current         = current
+            self.elens_length    = elens_length
+            self.voltage         = voltage
+            self.residual_kick_x   = residual_kick_x
+            self.residual_kick_y   = residual_kick_y
 
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(
@@ -132,50 +146,48 @@ class Elens(BeamElement):
                               voltage=self.voltage,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-Elens.XoStruct.extra_sources = [
-    _pkg_root.joinpath('beam_elements/elements_src/elens.h')]
-
-
-## Wire Element
 
 class Wire(BeamElement):
-    
+
     _xofields={
-               'wire_L_phy'  : xo.Float64,
-               'wire_L_int'  : xo.Float64,
-               'wire_current': xo.Float64,
-               'wire_xma'    : xo.Float64,
-               'wire_yma'    : xo.Float64,
+               'L_phy'  : xo.Float64,
+               'L_int'  : xo.Float64,
+               'current': xo.Float64,
+               'xma'    : xo.Float64,
+               'yma'    : xo.Float64,
               }
 
-    def __init__(self,  wire_L_phy   = None,
-                        wire_L_int   = None,
-                        wire_current = None,
-                        wire_xma     = None,
-                        wire_yma     = None,
+    _extra_c_sources = [
+        _pkg_root.joinpath('headers/constants.h'),
+        _pkg_root.joinpath('beam_elements/elements_src/wire.h'),
+    ]
+
+    def __init__(self,  L_phy   = 0,
+                        L_int   = 0,
+                        current = 0,
+                        xma     = 0,
+                        yma     = 0,
+                        _xobject = None,
                         **kwargs):
-        super().__init__(**kwargs)
-        self.wire_L_phy   = wire_L_phy
-        self.wire_L_int   = wire_L_int
-        self.wire_current = wire_current
-        self.wire_xma     = wire_xma
-        self.wire_yma     = wire_yma
+        if _xobject is not None:
+            super().__init__(_xobject=_xobject)
+        else:
+            super().__init__(**kwargs)
+            self.L_phy   = L_phy
+            self.L_int   = L_int
+            self.current = current
+            self.xma     = xma
+            self.yma     = yma
 
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(
-                              wire_L_phy   = self.wire_L_phy,
-                              wire_L_int   = self.wire_L_int,
-                              wire_current = -self.wire_current,
-                              wire_xma     = self.wire_xma,
-                              wire_yma     = self.wire_yma,
+                              L_phy   = self.L_phy,
+                              L_int   = self.L_int,
+                              current = -self.current,
+                              xma     = self.xma,
+                              yma     = self.yma,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-Wire.XoStruct.extra_sources = [
-     _pkg_root.joinpath('headers/constants.h'), 
-     _pkg_root.joinpath('beam_elements/elements_src/wire.h'), 
-]
-
-
 
 class SRotation(BeamElement):
     '''Beam element modeling an rotation of the reference system around the s axis. Parameters:
@@ -189,6 +201,9 @@ class SRotation(BeamElement):
         'sin_z': xo.Float64,
         }
 
+    _extra_c_sources = [
+        _pkg_root.joinpath('beam_elements/elements_src/srotation.h')]
+
     _store_in_to_dict = ['angle']
 
     def __init__(self, angle=0, **nargs):
@@ -201,23 +216,28 @@ class SRotation(BeamElement):
     def angle(self):
         return np.arctan2(self.sin_z, self.cos_z) * (180.0 / np.pi)
 
+    @angle.setter
+    def angle(self, value):
+        anglerad = value / 180 * np.pi
+        self.cos_z = np.cos(anglerad)
+        self.sin_z = np.sin(anglerad)
+
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(
                               angle=-self.angle,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-SRotation.XoStruct.extra_sources = [
-        _pkg_root.joinpath('beam_elements/elements_src/srotation.h')]
 
+class SynchrotronRadiationRecord(xo.HybridClass):
+    _xofields = {
+        '_index': RecordIndex,
+        'photon_energy': xo.Float64[:],
+        'at_element': xo.Int64[:],
+        'at_turn': xo.Int64[:],
+        'particle_id': xo.Int64[:],
+        'particle_delta': xo.Float64[:]
+        }
 
-def _update_bal_from_knl_ksl(knl, ksl, bal, context=None):
-    assert len(bal) == 2*len(knl) == 2*len(ksl)
-    idx = np.array([ii for ii in range(0, len(knl))])
-    inv_factorial = 1.0 / factorial(idx, exact=True)
-    if context is not None:
-        inv_factorial = context.nparray_to_context_array(inv_factorial)
-    bal[0::2] = knl * inv_factorial
-    bal[1::2] = ksl * inv_factorial
 
 class Multipole(BeamElement):
     '''Beam element modeling a thin magnetic multipole. Parameters:
@@ -225,7 +245,7 @@ class Multipole(BeamElement):
             - order [int]: Horizontal shift. Default is ``0``.
             - knl [m^-n, array]: Normalized integrated strength of the normal components.
             - ksl [m^-n, array]: Normalized integrated strength of the skew components.
-            - hxl [rad]: Rotation angle of the reference trajectoryin the horizzontal plane.
+            - hxl [rad]: Rotation angle of the reference trajectory in the horizontal plane.
             - hyl [rad]: Rotation angle of the reference trajectory in the vertical plane.
             - length [m]: Length of the originating thick multipole.
 
@@ -233,129 +253,75 @@ class Multipole(BeamElement):
 
     _xofields={
         'order': xo.Int64,
+        'inv_factorial_order': xo.Float64,
         'length': xo.Float64,
         'hxl': xo.Float64,
         'hyl': xo.Float64,
         'radiation_flag': xo.Int64,
-        'bal': xo.Float64[:],
+        'knl': xo.Float64[:],
+        'ksl': xo.Float64[:],
         }
 
-    _store_in_to_dict = ['knl', 'ksl']
+    _extra_c_sources = [
+        xp.general._pkg_root.joinpath('random_number_generator/rng_src/base_rng.h'),
+        xp.general._pkg_root.joinpath('random_number_generator/rng_src/local_particle_rng.h'),
+        _pkg_root.joinpath('headers/constants.h'),
+        _pkg_root.joinpath('headers/synrad_spectrum.h'),
+        _pkg_root.joinpath('beam_elements/elements_src/multipole.h')]
 
-    def __init__(self, order=None, knl=None, ksl=None, bal=None, **kwargs):
+    _internal_record_class = SynchrotronRadiationRecord
 
-        if bal is None and (
-            knl is not None or ksl is not None or order is not None
-        ):
-            if knl is None:
-                knl = []
-            if ksl is None:
-                ksl = []
-            if order is None:
-                order = 0
+    def __init__(self, order=None, knl=None, ksl=None, **kwargs):
 
-            n = max((order + 1), max(len(knl), len(ksl)))
-            assert n > 0
+        if '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
+            self.xoinitialize(**kwargs)
+            return
 
-            _knl = np.array(knl)
-            nknl = np.zeros(n, dtype=_knl.dtype)
-            nknl[: len(knl)] = knl
-            knl = nknl
-            del _knl
-            assert len(knl) == n
+        if order is None:
+            order = 0
 
-            _ksl = np.array(ksl)
-            nksl = np.zeros(n, dtype=_ksl.dtype)
-            nksl[: len(ksl)] = ksl
-            ksl = nksl
-            del _ksl
-            assert len(ksl) == n
+        if "bal" in kwargs.keys():
+            if not "knl" in kwargs.keys() or not "ksl" in kwargs.keys():
+                _bal = kwargs['bal']
+                idxes = np.array([ii for ii in range(0, len(_bal), 2)])
+                knl = [_bal[idx] * factorial(idx // 2, exact=True) for idx in idxes]
+                ksl = [_bal[idx + 1] * factorial(idx // 2, exact=True) for idx in idxes]
 
-            order = n - 1
-            bal = np.zeros(2 * order + 2)
+        len_knl = len(knl) if knl is not None else 0
+        len_ksl = len(ksl) if ksl is not None else 0
+        n = max((order + 1), max(len_knl, len_ksl))
+        assert n > 0
 
-            _update_bal_from_knl_ksl(knl, ksl, bal)
+        nknl = np.zeros(n, dtype=np.float64)
+        nksl = np.zeros(n, dtype=np.float64)
 
-            kwargs["bal"] = bal
-            kwargs["order"] = order
+        if knl is not None:
+            nknl[: len(knl)] = np.array(knl)
 
-        elif bal is not None and len(bal) > 0:
-            kwargs["bal"] = bal
-            kwargs["order"] = (len(bal) - 2) // 2
+        if ksl is not None:
+            nksl[: len(ksl)] = np.array(ksl)
+
+        order = n - 1
+
+        kwargs["knl"] = nknl
+        kwargs["ksl"] = nksl
+        kwargs["order"] = order
+        kwargs["inv_factorial_order"] = 1.0 / factorial(order, exact=True)
 
         self.xoinitialize(**kwargs)
 
-    @property
-    def knl(self):
-        bal_length = len(self.bal)
-        idxes = np.array([ii for ii in range(0, bal_length, 2)])
-        _bal = self._buffer.context.nparray_from_context_array(self.bal)
-        _knl = self._buffer.context.nparray_to_context_array(np.array(
-            [_bal[idx] * factorial(idx // 2, exact=True) for idx in idxes]))
-        return self._buffer.context.linked_array_type.from_array(
-                                        _knl,
-                                        mode='setitem_from_container',
-                                        container=self,
-                                        container_setitem_name='_knl_setitem')
-
-    @knl.setter
-    def knl(self, value):
-        self.knl[:] = value
-
-    def _knl_setitem(self, indx, val):
-        _knl = self.knl.copy()
-        _knl[indx] = val
-        _update_bal_from_knl_ksl(_knl, self.ksl, self.bal,
-                                 context=self._buffer.context)
-
-    @property
-    def ksl(self):
-        bal_length = len(self.bal)
-        idxes = np.array([ii for ii in range(0, bal_length, 2)])
-        _bal = self._buffer.context.nparray_from_context_array(self.bal)
-        _ksl = self._buffer.context.nparray_to_context_array(np.array(
-            [_bal[idx + 1] * factorial(idx // 2, exact=True) for idx in idxes]))
-        return self._buffer.context.linked_array_type.from_array(
-                                        _ksl,
-                                        mode='setitem_from_container',
-                                        container=self,
-                                        container_setitem_name='_ksl_setitem')
-
-    @ksl.setter
-    def ksl(self, value):
-        self.ksl[:] = value
-
-    def _ksl_setitem(self, indx, val):
-        _ksl = self.ksl.copy()
-        _ksl[indx] = val
-        _update_bal_from_knl_ksl(self.knl, _ksl, self.bal,
-                                 context=self._buffer.context)
-
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
+        ctx2np = self._buffer.context.nparray_from_context_array
         return self.__class__(
                               order=self.order,
                               length=-self.length,
                               hxl=-self.hxl,
                               hyl=-self.hyl,
                               radiation_flag=0, #TODO, I force radiation off for now
-                              bal=-self.bal, # TODO: maybe it can be made more efficient
+                              knl=-ctx2np(self.knl), # TODO: maybe it can be made more efficient
+                              ksl=-ctx2np(self.ksl), # TODO: maybe it can be made more efficient
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-Multipole.XoStruct.extra_sources = [
-    xp.general._pkg_root.joinpath('random_number_generator/rng_src/base_rng.h'),
-    xp.general._pkg_root.joinpath('random_number_generator/rng_src/local_particle_rng.h'),
-    _pkg_root.joinpath('headers/constants.h'),
-    _pkg_root.joinpath('headers/synrad_spectrum.h'),
-    _pkg_root.joinpath('beam_elements/elements_src/multipole.h')]
-
-def _update_phase_from_pn_ps(pn, ps, phase, context=None):
-    assert len(phase) == 2*len(pn) == 2*len(ps)
-    idx = np.array([ii for ii in range(0, len(pn))])
-    inv_factorial = 1.0 / factorial(idx, exact=True)
-    if context is not None:
-        inv_factorial = context.nparray_to_context_array(inv_factorial)
-    phase[0::2] = pn * inv_factorial
-    phase[1::2] = ps * inv_factorial
 
 class RFMultipole(BeamElement):
     '''Beam element modeling a thin modulated multipole, with strengths dependent on the z coordinate:
@@ -379,14 +345,19 @@ class RFMultipole(BeamElement):
 
     _xofields={
         'order': xo.Int64,
+        'inv_factorial_order': xo.Float64,
         'voltage': xo.Float64,
         'frequency': xo.Float64,
         'lag': xo.Float64,
-        'bal': xo.Float64[:],
-        'phase': xo.Float64[:],
+        'knl': xo.Float64[:],
+        'ksl': xo.Float64[:],
+        'pn': xo.Float64[:],
+        'ps': xo.Float64[:],
     }
 
-    _store_in_to_dict = ['knl', 'ksl', 'pn', 'ps']
+    _extra_c_sources = [
+        _pkg_root.joinpath('headers/constants.h'),
+        _pkg_root.joinpath('beam_elements/elements_src/rfmultipole.h')]
 
     def __init__(
         self,
@@ -395,158 +366,61 @@ class RFMultipole(BeamElement):
         ksl=None,
         pn=None,
         ps=None,
-        bal=None,
-        phase=None,
         **kwargs
     ):
 
+        if '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
+            self.xoinitialize(**kwargs)
+            return
+
         assert 'p' not in kwargs, "`p` in RF Multipole is not supported anymore"
 
-        if bal is None and (
-            knl is not None
-            or ksl is not None
-            or pn is not None
-            or ps is not None
-            or order is not None
-        ):
-            if knl is None:
-                knl = []
-            if ksl is None:
-                ksl = []
-            if pn is None:
-                pn = []
-            if ps is None:
-                ps = []
-            if order is None:
-                order = 0
+        if order is None:
+            order = 0
 
-            n = max((order + 1), max(len(knl), len(ksl), len(pn), len(ps)))
-            assert n > 0
-
-            _knl = np.array(knl)
-            nknl = np.zeros(n, dtype=_knl.dtype)
-            nknl[: len(knl)] = knl
-            knl = nknl
-            del _knl
-            assert len(knl) == n
-
-            _ksl = np.array(ksl)
-            nksl = np.zeros(n, dtype=_ksl.dtype)
-            nksl[: len(ksl)] = ksl
-            ksl = nksl
-            del _ksl
-            assert len(ksl) == n
-
-            _pn = np.array(pn)
-            npn = np.zeros(n, dtype=_pn.dtype)
-            npn[: len(pn)] = pn
-            pn = npn
-            del _pn
-            assert len(pn) == n
-
-            _ps = np.array(ps)
-            nps = np.zeros(n, dtype=_ps.dtype)
-            nps[: len(ps)] = ps
-            ps = nps
-            del _ps
-            assert len(ps) == n
-
-            order = n - 1
-            bal = np.zeros(2 * order + 2)
-            phase = np.zeros(2 * order + 2)
-
-            idx = np.array([ii for ii in range(0, len(knl))])
-            inv_factorial = 1.0 / factorial(idx, exact=True)
-            bal[0::2] = knl * inv_factorial
-            bal[1::2] = ksl * inv_factorial
-
-            phase[0::2] = pn
-            phase[1::2] = ps
-
-            kwargs["bal"] = bal
-            kwargs["phase"] = phase
-            kwargs["order"] = order
-
-        elif (
-            bal is not None
-            #and bal
-            and len(bal) >= 2
-            and ((len(bal) % 2) == 0)
-            and phase is not None
-            #and phase
-            and len(phase) >= 2
-            and ((len(phase) % 2) == 0)
-        ):
-            kwargs["bal"] = bal
-            kwargs["phase"] = phase
-            kwargs["order"] = (len(bal) - 2) / 2
-        elif '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
-            pass
-        else:
-            raise ValueError('RF Multipole Invalid input!')
+        if "bal" in kwargs.keys():
+            if not "knl" in kwargs.keys() or not "ksl" in kwargs.keys():
+                _bal = kwargs['bal']
+                idxes = np.array([ii for ii in range(0, len(_bal), 2)])
+                knl = [_bal[idx] * factorial(idx // 2, exact=True) for idx in idxes]
+                ksl = [_bal[idx + 1] * factorial(idx // 2, exact=True) for idx in idxes]
 
 
-        if '_xobject' in kwargs.keys() and kwargs['_xobject'] is not None:
-            super().__init__(**kwargs)
-        else:
-            temp_bal = kwargs["bal"]
-            temp_phase = kwargs["phase"]
+        len_knl = len(knl) if knl is not None else 0
+        len_ksl = len(ksl) if ksl is not None else 0
+        len_pn = len(pn) if pn is not None else 0
+        len_ps = len(ps) if ps is not None else 0
+        n = max((order + 1), max(len_knl, len_ksl, len_pn, len_ps))
+        assert n > 0
 
-            kwargs["bal"] = len(temp_bal)
-            kwargs["phase"] = len(temp_phase)
+        nknl = np.zeros(n, dtype=np.float64)
+        nksl = np.zeros(n, dtype=np.float64)
+        npn = np.zeros(n, dtype=np.float64)
+        nps = np.zeros(n, dtype=np.float64)
 
-            super().__init__(**kwargs)
+        if knl is not None:
+            nknl[: len(knl)] = np.array(knl)
 
-            ctx = self._buffer.context
-            self.bal[:] = ctx.nparray_to_context_array(temp_bal)
-            self.phase[:] = ctx.nparray_to_context_array(temp_phase)
+        if ksl is not None:
+            nksl[: len(ksl)] = np.array(ksl)
 
+        if pn is not None:
+            npn[: len(pn)] = np.array(pn)
 
-    @property
-    def pn(self):
-        phase_length = len(self.phase)
-        idxes = np.array([ii for ii in range(0, phase_length, 2)])
-        _phase = self._buffer.context.nparray_from_context_array(self.phase)
-        _pn = self._buffer.context.nparray_to_context_array(np.array(
-            [_phase[idx] * factorial(idx // 2, exact=True) for idx in idxes]))
-        return self._buffer.context.linked_array_type.from_array(
-                                        _pn,
-                                        mode='setitem_from_container',
-                                        container=self,
-                                        container_setitem_name='_pn_setitem')
+        if ps is not None:
+            nps[: len(ps)] = np.array(ps)
 
-    @pn.setter
-    def pn(self, value):
-        self.pn[:] = value
+        order = n - 1
 
-    def _pn_setitem(self, indx, val):
-        _pn = self.pn.copy()
-        _pn[indx] = val
-        _update_phase_from_pn_ps(_pn, self.ps, self.phase,
-                                 context=self._buffer.context)
+        kwargs["knl"] = nknl
+        kwargs["ksl"] = nksl
+        kwargs["pn"] = npn
+        kwargs["ps"] = nps
+        kwargs["order"] = order
+        #kwargs["inv_factorial_order"] = 1.0 / factorial(order, exact=True)
 
-    @property
-    def ps(self):
-        phase_length = len(self.phase)
-        idxes = np.array([ii for ii in range(0, phase_length, 2)])
-        _phase = self._buffer.context.nparray_from_context_array(self.phase)
-        _ps = self._buffer.context.nparray_to_context_array(np.array(
-            [_phase[idx + 1] * factorial(idx // 2, exact=True) for idx in idxes]))
-        return self._buffer.context.linked_array_type.from_array(
-                                        _ps,
-                                        mode='setitem_from_container',
-                                        container=self,
-                                        container_setitem_name='_ps_setitem')
+        self.xoinitialize(**kwargs)
 
-    @ps.setter
-    def ps(self, value):
-        self.ps[:] = value
-
-    def _ps_setitem(self, indx, val):
-        _ps = self.ps.copy()
-        _ps[indx] = val
-        _update_phase_from_pn_ps(self.knl, _ps, self.phase,
-                                 context=self._buffer.context)
 
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(
@@ -554,19 +428,12 @@ class RFMultipole(BeamElement):
                               voltage=-self.voltage,
                               frequency=self.frequency,
                               lag=self.lag,
-                              bal=[-bb for bb in self.bal], # TODO: maybe it can be made more efficient
-                              p = [pp for pp in self.phase],
+                              knl=-self.knl,
+                              ksl=-self.ksl,
+                              pn = self.pn,
+                              ps = self.ps,
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
-RFMultipole.knl = Multipole.knl
-RFMultipole.ksl = Multipole.ksl
-RFMultipole._knl_setitem = Multipole._knl_setitem
-RFMultipole._ksl_setitem = Multipole._ksl_setitem
-
-RFMultipole.XoStruct.extra_sources = [
-        _pkg_root.joinpath('headers/constants.h'),
-        _pkg_root.joinpath('beam_elements/elements_src/rfmultipole.h')]
-
 
 class DipoleEdge(BeamElement):
     '''Beam element modeling a dipole edge. Parameters:
@@ -581,9 +448,17 @@ class DipoleEdge(BeamElement):
     _xofields = {
             'r21': xo.Float64,
             'r43': xo.Float64,
+            'hgap': xo.Float64,
+            'h': xo.Float64,
+            'e1': xo.Float64,
+            'fint': xo.Float64,
             }
 
+    _extra_c_sources = [
+        _pkg_root.joinpath('beam_elements/elements_src/dipoleedge.h')]
+
     _store_in_to_dict = ['h', 'e1', 'hgap', 'fint']
+    _skip_in_to_dict = ['r21', 'r43']
 
     def __init__(
         self,
@@ -595,44 +470,36 @@ class DipoleEdge(BeamElement):
         fint=None,
         **kwargs
     ):
-        if r21 is None and r43 is None:
-            ZERO = np.float64(0.0)
-            if hgap is None:
-                hgap = ZERO
-            if h is None:
-                h = ZERO
-            if e1 is None:
-                e1 = ZERO
-            if fint is None:
-                fint = ZERO
 
-            # Check that the argument e1 is not too close to ( 2k + 1 ) * pi/2
-            # so that the cos in the denominator of the r43 calculation and
-            # the tan in the r21 calculations blow up
-            assert not np.isclose(np.absolute(np.cos(e1)), ZERO)
+        if r21 is not None or r43 is not None:
+            raise NotImplementedError(
+                "Please initialize using `h`, `e1`, `hgap` and `fint`")
 
-            corr = np.float64(2.0) * h * hgap * fint
-            r21 = h * np.tan(e1)
-            temp = corr / np.cos(e1) * (np.float64(1) + np.sin(e1) * np.sin(e1))
+        if hgap is None:
+            hgap = 0.
+        if h is None:
+            h = 0.
+        if e1 is None:
+            e1 = 0.
+        if fint is None:
+            fint = 0.
 
-            # again, the argument to the tan calculation should be limited
-            assert not np.isclose(np.absolute(np.cos(e1 - temp)), ZERO)
-            r43 = -h * np.tan(e1 - temp)
+        # Check that the argument e1 is not too close to ( 2k + 1 ) * pi/2
+        # so that the cos in the denominator of the r43 calculation and
+        # the tan in the r21 calculations blow up
+        assert not np.isclose(np.absolute(np.cos(e1)), 0)
 
-            self.h = h
-            self.e1 = e1
-            self.hgap = hgap
-            self.fint = fint
+        corr = np.float64(2.0) * h * hgap * fint
+        r21 = h * np.tan(e1)
+        temp = corr / np.cos(e1) * (np.float64(1) + np.sin(e1) * np.sin(e1))
+
+        # again, the argument to the tan calculation should be limited
+        assert not np.isclose(np.absolute(np.cos(e1 - temp)), 0)
+        r43 = -h * np.tan(e1 - temp)
+
+        super().__init__(h=h, hgap=hgap, e1=e1, fint=fint, r21=r21, r43=r43,
+                         **kwargs)
 
-        if r21 is not None and r43 is not None:
-            kwargs['r21'] = r21
-            kwargs['r43'] = r43
-            super().__init__(**kwargs)
-        else:
-            raise ValueError(
-                "DipoleEdge needs either coefficiants r21 and r43"
-                " or suitable values for h, e1, hgap, and fint provided"
-            )
 
     def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
         return self.__class__(
@@ -641,10 +508,6 @@ class DipoleEdge(BeamElement):
                               _context=_context, _buffer=_buffer, _offset=_offset)
 
 
-DipoleEdge.XoStruct.extra_sources = [
-        _pkg_root.joinpath('beam_elements/elements_src/dipoleedge.h')]
-
-
 class LinearTransferMatrix(BeamElement):
     _xofields={
         'no_detuning': xo.Int64,
@@ -683,8 +546,22 @@ class LinearTransferMatrix(BeamElement):
         'px_ref_1': xo.Float64,
         'y_ref_1': xo.Float64,
         'py_ref_1': xo.Float64,
+        'uncorrelated_rad_damping': xo.Int64,
+        'damping_factor_x':xo.Float64,
+        'damping_factor_y':xo.Float64,
+        'damping_factor_s':xo.Float64,
+        'uncorrelated_gauss_noise': xo.Int64,
+        'gauss_noise_ampl_x':xo.Float64,
+        'gauss_noise_ampl_y':xo.Float64,
+        'gauss_noise_ampl_s':xo.Float64,
         }
 
+    _extra_c_sources = [
+        xp.general._pkg_root.joinpath('random_number_generator/rng_src/base_rng.h'),
+        xp.general._pkg_root.joinpath('random_number_generator/rng_src/local_particle_rng.h'),
+        _pkg_root.joinpath('headers/constants.h'),
+        _pkg_root.joinpath('beam_elements/elements_src/lineartransfermatrix.h')]
+
     def __init__(self, Q_x=0, Q_y=0,
                      beta_x_0=1.0, beta_x_1=1.0, beta_y_0=1.0, beta_y_1=1.0,
                      alpha_x_0=0.0, alpha_x_1=0.0, alpha_y_0=0.0, alpha_y_1=0.0,
@@ -695,6 +572,9 @@ class LinearTransferMatrix(BeamElement):
                      energy_increment=0.0, energy_ref_increment=0.0,
                      x_ref_0 = 0.0, px_ref_0 = 0.0, x_ref_1 = 0.0, px_ref_1 = 0.0,
                      y_ref_0 = 0.0, py_ref_0 = 0.0, y_ref_1 = 0.0, py_ref_1 = 0.0,
+                     damping_rate_x = 0.0, damping_rate_y = 0.0, damping_rate_s = 0.0,
+                     equ_emit_x = 0.0, equ_emit_y = 0.0, equ_emit_s = 0.0,
+                     gauss_noise_ampl_x=0.0,gauss_noise_ampl_y=0.0,gauss_noise_ampl_s=0.0,
                      **nargs):
 
         if (chroma_x==0 and chroma_y==0
@@ -760,6 +640,35 @@ class LinearTransferMatrix(BeamElement):
         # acceleration without change of reference momentum
         nargs['energy_increment'] = energy_increment
 
+        if damping_rate_x < 0.0 or damping_rate_y < 0.0 or damping_rate_s < 0.0:
+            raise ValueError('Damping rates cannot be negative')
+        if damping_rate_x > 0.0 or damping_rate_y > 0.0 or damping_rate_s > 0.0:
+            nargs['uncorrelated_rad_damping'] = True
+            nargs['damping_factor_x'] = 1.0-damping_rate_x/2.0
+            nargs['damping_factor_y'] = 1.0-damping_rate_y/2.0
+            nargs['damping_factor_s'] = 1.0-damping_rate_s/2.0
+        else:
+            nargs['uncorrelated_rad_damping'] = False
+
+        if equ_emit_x < 0.0 or equ_emit_y < 0.0 or equ_emit_s < 0.0:
+            raise ValueError('Equilibrium emittances cannot be negative')
+        nargs['uncorrelated_gauss_noise'] = False
+        nargs['gauss_noise_ampl_x'] = 0.0
+        nargs['gauss_noise_ampl_y'] = 0.0
+        nargs['gauss_noise_ampl_s'] = 0.0
+        if equ_emit_x > 0.0 or equ_emit_y > 0.0 or equ_emit_s > 0.0:
+            nargs['uncorrelated_gauss_noise'] = True
+            nargs['gauss_noise_ampl_x'] = np.sqrt(2.0*equ_emit_x*damping_rate_x/beta_x_1)
+            nargs['gauss_noise_ampl_y'] = np.sqrt(2.0*equ_emit_y*damping_rate_y/beta_y_1)
+            nargs['gauss_noise_ampl_s'] = np.sqrt(2.0*equ_emit_s*damping_rate_s/beta_s)
+
+        if gauss_noise_ampl_x < 0.0 or gauss_noise_ampl_y < 0.0 or gauss_noise_ampl_s < 0.0:
+            raise ValueError('Gaussian noise amplitude cannot be negative')
+        if gauss_noise_ampl_x > 0.0 or gauss_noise_ampl_y > 0.0 or gauss_noise_ampl_s > 0.0:
+            nargs['uncorrelated_gauss_noise'] = True
+            nargs['gauss_noise_ampl_x'] = np.sqrt(nargs['gauss_noise_ampl_x']**2+gauss_noise_ampl_x**2)
+            nargs['gauss_noise_ampl_y'] = np.sqrt(nargs['gauss_noise_ampl_y']**2+gauss_noise_ampl_y**2)
+            nargs['gauss_noise_ampl_s'] = np.sqrt(nargs['gauss_noise_ampl_s']**2+gauss_noise_ampl_s**2)
 
         super().__init__(**nargs)
 
@@ -775,22 +684,44 @@ class LinearTransferMatrix(BeamElement):
     def beta_y_1(self):
         return self.beta_prod_y*self.beta_ratio_y
 
-LinearTransferMatrix.XoStruct.extra_sources = [
-        _pkg_root.joinpath('headers/constants.h'),
-        _pkg_root.joinpath('beam_elements/elements_src/lineartransfermatrix.h')]
 
-class EnergyChange(BeamElement):
+
+class FirstOrderTaylorMap(BeamElement):
+    isthick = True
+
     _xofields={
-        'energy_ref_increment': xo.Float64,
-        'energy_increment': xo.Float64
-        }
+        'radiation_flag': xo.Int64,
+        'length': xo.Float64,
+        'm0': xo.Float64[6],
+        'm1': xo.Float64[6,6]}
 
-    def __init__(self, energy_increment=0.0,energy_ref_increment=0.0, **nargs):
-        nargs['energy_ref_increment']=energy_ref_increment # acceleration with change of reference momentum (e.g. ramp)
-        nargs['energy_increment']=energy_increment # acceleration without change of reference momentum (e.g. compensation of energy loss)
+    _extra_c_sources = [
+        xp.general._pkg_root.joinpath('random_number_generator/rng_src/base_rng.h'),
+        xp.general._pkg_root.joinpath('random_number_generator/rng_src/local_particle_rng.h'),
+        _pkg_root.joinpath('headers/constants.h'),
+        _pkg_root.joinpath('headers/synrad_spectrum.h'),
+        _pkg_root.joinpath('beam_elements/elements_src/firstordertaylormap.h')]
+
+    _internal_record_class = SynchrotronRadiationRecord # not functional,
+    # included for compatibility with Multipole
+
+    def __init__(self, length = 0.0, m0 = None, m1 = None,radiation_flag=0,**nargs):
+        nargs['radiation_flag'] = radiation_flag
+        nargs['length'] = length
+        if m0 is None:
+            nargs['m0'] = np.zeros(6,dtype=np.float64)
+        else:
+            if len(np.shape(m0)) == 1 and np.shape(m0)[0] == 6:
+                nargs['m0'] = m0
+            else:
+                raise ValueError(f'Wrong shape for m0: {np.shape(m0)}')
+        if m1 is None:
+            nargs['m1'] = np.eye(6,dtype=np.float64)
+        else:
+            if len(np.shape(m1)) == 2 and np.shape(m1)[0] == 6 and np.shape(m1)[1] == 6:
+                nargs['m1'] = m1
+            else:
+                raise ValueError(f'Wrong shape for m1: {np.shape(m1)}')
         super().__init__(**nargs)
 
-EnergyChange.XoStruct.extra_sources = [
-        _pkg_root.joinpath('beam_elements/elements_src/energychange.h')]
-
 
