@@ -1,27 +1,31 @@
-import json
+# copyright ############################### #
+# This file is part of the Xtrack Package.  #
+# Copyright (c) CERN, 2021.                 #
+# ######################################### #
+
 import math
 import logging
 from copy import deepcopy
+from pprint import pp
 
 import numpy as np
 
 import xobjects as xo
 import xpart as xp
 
-from .loader_sixtrack import _expand_struct
-from .loader_mad import madx_sequence_to_xtrack_line
+from .mad_loader import MadLoader
 from .beam_elements import element_classes, Multipole
 from . import beam_elements
 from .beam_elements import Drift
 
-
 log=logging.getLogger(__name__)
 
 def mk_class_namespace(extra_classes):
     try:
-       import xfields as xf
-       all_classes= element_classes + xf.element_classes + extra_classes
+        import xfields as xf
+        all_classes = element_classes + xf.element_classes + extra_classes + (Line,)
     except ImportError:
+        all_classes = element_classes + extra_classes
         log.warning("Xfields not installed correctly")
 
     out=AttrDict()
@@ -58,16 +62,24 @@ class Line:
     def from_dict(cls, dct, _context=None, _buffer=None, classes=()):
         class_dict=mk_class_namespace(classes)
 
-        _buffer, _ =xo.get_a_buffer(size=8,context=_context, buffer=_buffer)
+        _buffer, _ = xo.get_a_buffer(size=8,context=_context, buffer=_buffer)
+        num_elements = len(dct['element_names'])
         elements = []
-        for el in dct["elements"]:
+        for ii, el in enumerate(dct["elements"]):
+
+            if ii % 100 == 0:
+                print(
+                    f'Loading line from dict: {round(ii/num_elements*100):2d}%  ',
+                    end="\r", flush=True)
+
             eltype = class_dict[el["__class__"]]
             eldct=el.copy()
             del eldct['__class__']
-            if hasattr(eltype,'XoStruct'):
+            if hasattr(eltype,'_XoStruct'):
                newel = eltype.from_dict(eldct,_buffer=_buffer)
             else:
                newel = eltype.from_dict(eldct)
+
             elements.append(newel)
 
         self = cls(elements=elements, element_names=dct['element_names'])
@@ -82,27 +94,18 @@ class Line:
             for kk in self._var_management['data'].keys():
                 self._var_management['data'][kk].update(
                                             dct['_var_management_data'][kk])
-            manager.reload(dct['_var_manager'])
+            manager.load(dct['_var_manager'])
+
+        print('Done loading line from dict.           ')
 
         return self
 
     @classmethod
     def from_sixinput(cls, sixinput, classes=()):
-        class_dict=mk_class_namespace(classes)
-
-        line_data, rest, iconv = _expand_struct(sixinput, convert=class_dict)
-
-        ele_names = [dd[0] for dd in line_data]
-        elements = [dd[2] for dd in line_data]
-
-        line = cls(elements=elements, element_names=ele_names)
-
-        other_info = {}
-        other_info["rest"] = rest
-        other_info["iconv"] = iconv
-
-        line.other_info = other_info
-
+        log.warning("\n"
+            "WARNING: xtrack.Line.from_sixinput(sixinput) will be removed in furure versions.\n"
+            "Please use sixinput.generate_xtrack_line()\n")
+        line = sixinput.generate_xtrack_line(classes=classes)
         return line
 
     @classmethod
@@ -110,28 +113,35 @@ class Line:
         cls,
         sequence,
         classes=(),
-        ignored_madtypes=[],
+        ignored_madtypes=(),
         exact_drift=False,
-        drift_threshold=1e-6,
+#        drift_threshold=1e-6, # not used anymore with expanded sequences
         deferred_expressions=False,
         install_apertures=False,
         apply_madx_errors=False,
+        skip_markers=False,
+        merge_drifts=False,
+        merge_multipoles=False,
     ):
 
-        class_dict=mk_class_namespace(classes)
+        if not (exact_drift is False):
+            raise NotImplementedError("Exact drifts not implemented yet")
 
-        line = madx_sequence_to_xtrack_line(
-            sequence,
-            class_dict,
-            ignored_madtypes=ignored_madtypes,
-            exact_drift=exact_drift,
-            drift_threshold=drift_threshold,
-            install_apertures=install_apertures,
-            deferred_expressions=deferred_expressions)
-
-        if apply_madx_errors:
-            line._apply_madx_errors(sequence)
+        class_namespace=mk_class_namespace(classes)
 
+        loader = MadLoader(sequence,
+            classes=class_namespace,
+            ignore_madtypes=ignored_madtypes,
+            exact_drift=False,
+            enable_errors=apply_madx_errors,
+            enable_apertures=install_apertures,
+            enable_expressions=deferred_expressions,
+            skip_markers=skip_markers,
+            merge_drifts=merge_drifts,
+            merge_multipoles=merge_multipoles,
+            error_table=None,  # not implemented yet
+            )
+        line=loader.make_line()
         return line
 
     def _init_var_management(self):
@@ -140,7 +150,8 @@ class Line:
         import xdeps as xd
 
         # Extract globals values from madx
-        _var_values=defaultdict(lambda :0)
+        _var_values = defaultdict(lambda :0)
+        _var_values.default_factory = None
 
         _ref_manager = manager=xd.Manager()
         _vref=manager.ref(_var_values,'vars')
@@ -200,11 +211,29 @@ class Line:
 
         self._var_management = None
         self._needs_rng = False
+        self.tracker = None
+
+    def build_tracker(self, **kwargs):
+        assert self.tracker is None, 'The line already has an associated tracker'
+        import xtrack as xt # avoid circular import
+        self.tracker = xt.Tracker(line=self, **kwargs)
+        return self.tracker
+
 
     @property
     def elements(self):
         return tuple([self.element_dict[nn] for nn in self.element_names])
 
+    def __getitem__(self, ii):
+        if isinstance(ii, str):
+            return self.element_dict.__getitem__(ii)
+        else:
+            names = self.element_names.__getitem__(ii)
+            if isinstance(names, str):
+                return self.element_dict.__getitem__(names)
+            else:
+                return [self.element_dict[nn] for nn in names]
+
     def filter_elements(self, mask=None, exclude_types_starting_with=None):
 
         if mask is None:
@@ -222,12 +251,16 @@ class Line:
                 new_elements.append(ee)
             else:
                 if _is_thick(ee) and not _is_drift(ee):
-                    new_elements.append(Drift(length==ee.length))
+                    new_elements.append(Drift(length=ee.length))
                 else:
                     new_elements.append(Drift(length=0))
 
-        return self.__class__(elements=new_elements,
+        new_line = self.__class__(elements=new_elements,
                               element_names=self.element_names)
+        if self.particle_ref is not None:
+            new_line.particle_ref = self.particle_ref.copy()
+
+        return new_line
 
     def cycle(self, index_first_element=None, name_first_element=None):
 
@@ -271,6 +304,11 @@ class Line:
     def _freeze(self):
         self.element_names = tuple(self.element_names)
 
+    def unfreeze(self):
+        self.element_names = list(self.element_names)
+        if hasattr(self, 'tracker') and self.tracker is not None:
+            self.tracker._invalidate()
+
     def _frozen_check(self):
         if isinstance(self.element_names, tuple):
             raise ValueError(
@@ -290,10 +328,32 @@ class Line:
             out['_var_manager'] = self._var_management['manager'].dump()
         return out
 
+    def to_pandas(self):
+        elements = self.elements
+        s_elements = np.array(self.get_s_elements())
+        element_types = list(map(lambda e: e.__class__.__name__, elements))
+        isthick = np.array(list(map(_is_thick, elements)))
+
+        import pandas as pd
+
+        elements_df = pd.DataFrame({
+            'element_type': element_types,
+            's': s_elements,
+            'name': self.element_names,
+            'isthick': isthick,
+            'element': elements
+        })
+        return elements_df
+
     def copy(self):
         return self.__class__.from_dict(self.to_dict())
 
-    def insert_element(self, index=None, element=None, name=None, at_s=None):
+    def insert_element(self, index=None, element=None, name=None, at_s=None,
+                       s_tol=1e-6):
+
+        if isinstance(index, str):
+            assert index in self.element_names
+            index = self.element_names.index(index)
 
         assert name is not None
         if element is None:
@@ -309,6 +369,13 @@ class Line:
 
         if at_s is not None:
             s_vect_upstream = np.array(self.get_s_position(mode='upstream'))
+
+            if not _is_thick(element) or np.abs(element.length)==0:
+                i_closest = np.argmin(np.abs(s_vect_upstream - at_s))
+                if np.abs(s_vect_upstream[i_closest] - at_s) < s_tol:
+                    return self.insert_element(index=i_closest,
+                                            element=element, name=name)
+
             s_vect_downstream = np.array(self.get_s_position(mode='downstream'))
 
             s_start_ele = at_s
@@ -320,13 +387,14 @@ class Line:
                     return self.insert_element(index=i_first_drift_to_cut,
                                               element=element, name=name)
 
-            if _is_thick(element):
+            if _is_thick(element) and np.abs(element.length)>0:
                 s_end_ele = at_s + element.length
             else:
                 s_end_ele = s_start_ele
 
             i_last_drift_to_cut = np.where(s_vect_upstream < s_end_ele)[0][-1]
-            assert i_first_drift_to_cut <= i_last_drift_to_cut
+            if _is_thick(element) and element.length > 0:
+                assert i_first_drift_to_cut <= i_last_drift_to_cut
             name_first_drift_to_cut = self.element_names[i_first_drift_to_cut]
             name_last_drift_to_cut = self.element_names[i_last_drift_to_cut]
             first_drift_to_cut = self.element_dict[name_first_drift_to_cut]
@@ -336,7 +404,9 @@ class Line:
             assert _is_drift(last_drift_to_cut)
 
             for ii in range(i_first_drift_to_cut, i_last_drift_to_cut+1):
-                if not _is_drift(self.element_dict[self.element_names[ii]]):
+                e_to_replace = self.element_dict[self.element_names[ii]]
+                if (not _is_drift(e_to_replace) and
+                    not e_to_replace.__class__.__name__.startswith('Limit')):
                     raise ValueError('Cannot replace active element '
                                         f'{self.element_names[ii]}')
 
@@ -377,7 +447,7 @@ class Line:
                     self.element_names.insert(i_insert, nn)
 
         else:
-            if _is_thick(element):
+            if _is_thick(element) and np.abs(element.length)>0:
                 raise NotImplementedError('use `at_s` to insert thick elements')
             assert name not in self.element_dict.keys()
             self.element_dict[name] = element
@@ -479,11 +549,10 @@ class Line:
                 continue
 
             if _is_drift(ee):
-                prev_ee = newline.elements[-1]
                 prev_nn = newline.element_names[-1]
+                prev_ee = newline.element_dict[prev_nn]
                 if _is_drift(prev_ee):
                     prev_ee.length += ee.length
-                    prev_nn += ('_' + nn)
                     newline.element_names[-1] = prev_nn
                 else:
                     newline.append_element(ee, nn)
@@ -500,7 +569,7 @@ class Line:
     def merge_consecutive_multipoles(self, inplace=False):
 
         self._frozen_check()
-        if hasattr(self, '_var_management'):
+        if self._var_management is not None:
             raise NotImplementedError('`merge_consecutive_multipoles` not'
                                       ' available when deferred expressions are'
                                       ' used')
@@ -566,189 +635,86 @@ class Line:
 
         return elements, names
 
-    def _find_element_ids(self, element_name):
-        """Find element_name in this Line instance's
-        self.elements_name list. Assumes the names are unique.
+    def check_aperture(self):
 
-        Return index before and after the element, taking into account
-        attached _aperture instances (LimitRect, LimitEllipse, ...)
-        which would follow the element occurrence in the list.
+        elements_df = self.to_pandas()
 
-        Raises IndexError if element_name not found in this Line.
-        """
-        # will raise error if element not present:
-        idx_el = self.element_names.index(element_name)
-        try:
-            # if aperture marker is present
-            idx_after_el = self.element_names.index(element_name + "_aperture") + 1
-        except ValueError:
-            # if aperture marker is not present
-            idx_after_el = idx_el + 1
-        return idx_el, idx_after_el
+        elements_df['is_aperture'] = elements_df.element_type.map(lambda s: s.startswith('Limit'))
+        elements_df['i_aperture_upstream'] = np.nan
+        elements_df['s_aperture_upstream'] = np.nan
+        elements_df['i_aperture_downstream'] = np.nan
+        elements_df['s_aperture_downstream'] = np.nan
 
-    def _add_offset_error_to(self, element_name, dx=0, dy=0):
-        idx_el, idx_after_el = self._find_element_ids(element_name)
-        xyshift = beam_elements.XYShift(dx=dx, dy=dy)
-        inv_xyshift = beam_elements.XYShift(dx=-dx, dy=-dy)
-        self.insert_element(idx_el, xyshift, element_name + "_offset_in")
-        self.insert_element(
-            idx_after_el + 1, inv_xyshift, element_name + "_offset_out"
-        )
+        num_elements = len(self.element_names)
 
-    def _add_aperture_offset_error_to(self, element_name, arex=0, arey=0):
-        idx_el, idx_after_el = self._find_element_ids(element_name)
-        idx_el_aper = idx_after_el - 1
-        if not self.element_names[idx_el_aper] == element_name + "_aperture":
-            # it is allowed to provide arex/arey without providing an aperture
-            print('Info: Element', element_name, ': arex/y provided without aperture -> arex/y ignored')
-            return
-        xyshift = beam_elements.XYShift(dx=arex, dy=arey)
-        inv_xyshift = beam_elements.XYShift(dx=-arex, dy=-arey)
-        self.insert_element(idx_el_aper, xyshift, element_name + "_aperture_offset_in")
-        self.insert_element(
-            idx_after_el + 1, inv_xyshift, element_name + "_aperture_offset_out"
-        )
+        i_prev_aperture = elements_df[elements_df['is_aperture']].index[0]
+        i_next_aperture = 0
 
-    def _add_tilt_error_to(self, element_name, angle):
-        '''Alignment error of transverse rotation around s-axis.
-        The element corresponding to the given `element_name`
-        gets wrapped by SRotation elements with rotation angle
-        `angle`.
+        for iee in range(i_prev_aperture, num_elements):
 
-        In the case of a thin dipole component, the corresponding
-        curvature terms in the Multipole (hxl and hyl) are rotated
-        by `angle` as well.
-        '''
-        idx_el, idx_after_el = self._find_element_ids(element_name)
-        element = self.elements[self.element_names.index(element_name)]
-        if isinstance(element, beam_elements.Multipole) and (
-                element.hxl or element.hyl):
-            dpsi = angle * deg2rad
+            if iee % 100 == 0:
+                print(
+                    f'Checking aperture: {round(iee/num_elements*100):2d}%  ',
+                    end="\r", flush=True)
 
-            hxl0 = element.hxl
-            hyl0 = element.hyl
+            if elements_df.loc[iee, 'element_type'] == 'Drift':
+                continue
 
-            hxl1 = hxl0 * np.cos(dpsi) - hyl0 * np.sin(dpsi)
-            hyl1 = hxl0 * np.sin(dpsi) + hyl0 * np.cos(dpsi)
+            if elements_df.loc[iee, 'element_type'] == 'XYShift':
+                continue
 
-            element.hxl = hxl1
-            element.hyl = hyl1
-        srot = beam_elements.SRotation(angle=angle)
-        inv_srot = beam_elements.SRotation(angle=-angle)
-        self.insert_element(idx_el, srot, element_name + "_tilt_in")
-        self.insert_element(idx_after_el + 1, inv_srot, element_name + "_tilt_out")
+            if elements_df.loc[iee, 'element_type'] == 'SRotation':
+                continue
 
-    def _add_multipole_error_to(self, element_name, knl=[], ksl=[]):
-        # will raise error if element not present:
-        assert element_name in self.element_names
-        element_index = self.element_names.index(element_name)
-        element = self.elements[element_index]
+            if elements_df.loc[iee, 'is_aperture']:
+                i_prev_aperture = iee
+                continue
 
-        new_order = max([len(knl), len(ksl), len(element.knl), len(element.ksl)])
-        new_knl = new_order*[0]
-        new_ksl = new_order*[0]
+            if i_next_aperture < iee:
+                for ii in range(iee, num_elements):
+                    if elements_df.loc[ii, 'is_aperture']:
+                        i_next_aperture = ii
+                        break
 
-        # Original strengths
-        for ii, vv in enumerate(element.knl):
-            new_knl[ii] += element.knl[ii]
-        for ii, vv in enumerate(element.ksl):
-            new_ksl[ii] += element.ksl[ii]
+            elements_df.at[iee, 'i_aperture_upstream'] = i_prev_aperture
+            elements_df.at[iee, 'i_aperture_downstream'] = i_next_aperture
 
-        # Errors
-        for ii, vv in enumerate(knl):
-            new_knl[ii] += knl[ii]
-        for ii, vv in enumerate(ksl):
-            new_ksl[ii] += ksl[ii]
+            elements_df.at[iee, 's_aperture_upstream'] = elements_df.loc[i_prev_aperture, 's']
+            elements_df.at[iee, 's_aperture_downstream'] = elements_df.loc[i_next_aperture, 's']
 
-        new_element = Multipole(knl=new_knl, ksl=new_ksl,
-                length=element.length, hxl=element.hxl,
-                hyl=element.hyl, radiation_flag=element.radiation_flag)
+        # Check for elements missing aperture upstream
+        elements_df['misses_aperture_upstream'] = ((elements_df['s_aperture_upstream'] != elements_df['s'])
+            & ~(np.isnan(elements_df['i_aperture_upstream'])))
 
-        self.element_dict[element_name] = new_element
+        # Check for elements missing aperture downstream
+        s_downstream = elements_df.s.copy()
+        df_thick_to_check = elements_df[elements_df['isthick'] & ~(elements_df.i_aperture_upstream.isna())].copy()
+        s_downstream.loc[df_thick_to_check.index] += np.array([ee.length for ee in df_thick_to_check.element])
+        elements_df['misses_aperture_downstream'] = (
+            (np.abs(elements_df['s_aperture_downstream'] - s_downstream) > 1e-6)
+            & ~(np.isnan(elements_df['i_aperture_upstream'])))
 
-        # Handle deferred expressions
-        if self._var_management is not None:
-            lref = self._var_management['lref']
-            manager = self._var_management['manager']
-            for ii in range(min([len(knl), len(element.knl)])):
-                lref[element_name].knl[ii] += knl[ii]
+        # Flag problems
+        elements_df['has_aperture_problem'] = (
+            elements_df['misses_aperture_upstream'] | (
+                elements_df['isthick'] & elements_df['misses_aperture_downstream']))
 
-            for ii in range(min([len(ksl), len(element.ksl)])):
-                lref[element_name].ksl[ii] += ksl[ii]
+        print('Done checking aperture.           ')
 
-    def _apply_madx_errors(self, madx_sequence):
-        """Applies errors from MAD-X sequence to existing
-        elements in this Line instance.
+        # Identify issues with apertures associate with thin elements
+        df_thin_missing_aper = elements_df[elements_df['misses_aperture_upstream'] & ~elements_df['isthick']]
+        print(f'{len(df_thin_missing_aper)} thin elements miss associated aperture (upstream):')
+        pp(list(df_thin_missing_aper.name))
 
-        Return names of MAD-X elements with existing align_errors
-        or field_errors which were not found in the elements of
-        this Line instance (and thus not treated).
+        # Identify issues with apertures associate with thin elements
+        df_thick_missing_aper = elements_df[
+            (elements_df['misses_aperture_upstream'] | elements_df['misses_aperture_downstream'])
+            & elements_df['isthick']]
+        print(f'{len(df_thick_missing_aper)} thick elements miss associated aperture (upstream or downstream):')
+        pp(list(df_thick_missing_aper.name))
 
-        Example via cpymad:
-            madx = cpymad.madx.Madx()
+        return elements_df
 
-            # (...set up lattice and errors in cpymad...)
-
-            seq = madx.sequence.some_lattice
-            line = Line.from_madx_sequence(
-                                    seq,
-                                    apply_madx_errors=True
-                              )
-        """
-        elements_not_found = []
-        for element, element_name in zip(
-                madx_sequence.expanded_elements,
-                madx_sequence.expanded_element_names()
-        ):
-            if element_name not in self.element_names:
-                if element.align_errors or element.field_errors:
-                    elements_not_found.append(element_name)
-                    continue
-
-            if element.align_errors:
-                # add offset
-                dx = element.align_errors.dx
-                dy = element.align_errors.dy
-                if dx or dy:
-                    self._add_offset_error_to(element_name, dx, dy)
-
-                # add tilt
-                dpsi = element.align_errors.dpsi
-                if dpsi:
-                    self._add_tilt_error_to(element_name, angle=dpsi / deg2rad)
-
-                # add aperture-only offset
-                arex = element.align_errors.arex
-                arey = element.align_errors.arey
-                if arex or arey:
-                    self._add_aperture_offset_error_to(element_name, arex, arey)
-
-                # check for errors which cannot be treated yet:
-                #for error_type in dir(element.align_errors):
-                 #   if not error_type[0] == '_' and \
-                  #          error_type not in ['dx', 'dy', 'dpsi', 'arex',
-                   #                            'arey', 'count', 'index']:
-                        #print(
-                        #    f'Warning: MAD-X error type "{error_type}"'
-                        #    " not implemented yet."
-                        #)
-
-            if element.field_errors:
-                # add multipole error
-                if any(element.field_errors.dkn) or \
-                            any(element.field_errors.dks):
-                    knl = element.field_errors.dkn
-                    ksl = element.field_errors.dks
-                    on=np.where(knl)[0]
-                    os=np.where(ksl)[0]
-                    on = on[-1] if len(on)>0 else 0
-                    os = os[-1] if len(os)>0 else 0
-                    oo = max(os,on)+1
-                    knl = knl[:oo]  # delete trailing zeros
-                    ksl = ksl[:oo]  # to keep order low
-                    self._add_multipole_error_to(element_name, knl, ksl)
-
-        return elements_not_found
 
 mathfunctions = type('math', (), {})
 mathfunctions.sqrt=math.sqrt
