@@ -1,3 +1,8 @@
+# copyright ############################### #
+# This file is part of the Xtrack Package.  #
+# Copyright (c) CERN, 2021.                 #
+# ######################################### #
+
 import numpy as np
 from scipy.spatial import ConvexHull
 
@@ -11,15 +16,23 @@ import logging
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.StreamHandler())
 
+_default_allowed_backtrack_types = [Drift, SRotation, XYShift]
+
 class LossLocationRefinement:
 
     def __init__(self, tracker, backtracker=None,
                  n_theta=None, r_max=None, dr=None, ds=None,
-                 save_refine_trackers=False):
+                 save_refine_trackers=False,
+                 allowed_backtrack_types=()):
+
+        allowed_backtrack_types = tuple(set(allowed_backtrack_types).union(
+                                        set(_default_allowed_backtrack_types)))
 
         if tracker.iscollective:
+            self._original_tracker = tracker
             self.tracker = tracker._supertracker
         else:
+            self._original_tracker = tracker
             self.tracker = tracker
 
         self._context = self.tracker._line_frozen._buffer.context
@@ -36,11 +49,13 @@ class LossLocationRefinement:
         # Build track kernel with all elements + polygon
         trk_gen = Tracker(_buffer=self.tracker._line_frozen._buffer,
                 line=Line(
-                    elements=self.tracker._line_frozen.elements + (temp_poly,)))
+                    elements=self.tracker._line_frozen.elements + (temp_poly,)),
+                    global_xy_limit=tracker.global_xy_limit)
         self._trk_gen = trk_gen
 
         if backtracker is None:
-            backtracker = self.tracker.get_backtracker(_context=self._context)
+            backtracker = self.tracker.get_backtracker(_context=self._context,
+                                                       global_xy_limit=None)
         self.backtracker = backtracker
 
         self.i_apertures, self.apertures = find_apertures(self.tracker)
@@ -53,6 +68,7 @@ class LossLocationRefinement:
         self.r_max = r_max
         self.dr = dr
         self.ds = ds
+        self.allowed_backtrack_types = allowed_backtrack_types
 
     def refine_loss_location(self, particles, i_apertures=None):
 
@@ -104,9 +120,11 @@ class LossLocationRefinement:
                                       self.n_theta, self.r_max, self.dr, self.ds,
                                       _trk_gen=self._trk_gen)
 
+                interp_tracker._original_tracker = self._original_tracker
                 part_refine = refine_loss_location_single_aperture(
-                            particles,i_aper_1, i_start_thin_0,
-                            self.backtracker, interp_tracker, inplace=True)
+                        particles,i_aper_1, i_start_thin_0,
+                        self.backtracker, interp_tracker, inplace=True,
+                        allowed_backtrack_types=self.allowed_backtrack_types)
 
                 if self.save_refine_trackers:
                     interp_tracker.i_start_thin_0 = i_start_thin_0
@@ -157,27 +175,46 @@ def find_apertures(tracker):
     return i_apertures, apertures
 
 def refine_loss_location_single_aperture(particles, i_aper_1, i_start_thin_0,
-                                         backtracker, interp_tracker,
-                                         inplace=True):
+                    backtracker, interp_tracker,
+                    inplace=True,
+                    allowed_backtrack_types=_default_allowed_backtrack_types):
 
     mask_part = (particles.state == 0) & (particles.at_element == i_aper_1)
+
     part_refine = xp.Particles(
                     p0c=particles.p0c[mask_part],
+                    mass0=particles.mass0,
+                    q0=particles.q0,
                     x=particles.x[mask_part],
                     px=particles.px[mask_part],
                     y=particles.y[mask_part],
                     py=particles.py[mask_part],
                     zeta=particles.zeta[mask_part],
                     delta=particles.delta[mask_part],
-                    s=particles.s[mask_part])
+                    s=particles.s[mask_part],
+                    chi=particles.chi[mask_part],
+                    charge_ratio=particles.charge_ratio[mask_part])
     n_backtrack = i_aper_1 - (i_start_thin_0+1)
-    num_elements = len(backtracker.line.elements)
+    num_elements = len(backtracker.line.element_names)
     i_start_backtrack = num_elements-i_aper_1
+
+    # Check that we are not backtracking through element types that are not allowed
+    for nn in interp_tracker._original_tracker.line.element_names[
+                                             i_aper_1 - n_backtrack : i_aper_1]:
+        ee = interp_tracker._original_tracker.line.element_dict[nn]
+        if not isinstance(ee, tuple(allowed_backtrack_types)):
+            if (hasattr(ee, 'skip_in_loss_location_refinement')
+                    and ee.skip_in_loss_location_refinement):
+                return 'skipped'
+            raise TypeError(
+                f'Cannot backtrack through element {nn} of type '
+                f'{ee.__class__.__name__}')
+
     backtracker.track(part_refine, ele_start=i_start_backtrack,
                       num_elements = n_backtrack)
     # Just for check
-    elem_backtrack = backtracker.line.elements[
-                        i_start_backtrack:i_start_backtrack + n_backtrack]
+    # elem_backtrack = backtracker.line.elements[
+    #                     i_start_backtrack:i_start_backtrack + n_backtrack]
 
     # Track with extra apertures
     interp_tracker.track(part_refine)
@@ -185,7 +222,6 @@ def refine_loss_location_single_aperture(particles, i_aper_1, i_start_thin_0,
     # We verified that they are really at the edge. Their coordinates
     # correspond to the end fo the short line, which is correct
 
-
     if inplace:
         indx_sorted = np.argsort(part_refine.particle_id)
         with particles._bypass_linked_vars():
@@ -196,7 +232,7 @@ def refine_loss_location_single_aperture(particles, i_aper_1, i_start_thin_0,
             particles.zeta[mask_part] = part_refine.zeta[indx_sorted]
             particles.s[mask_part] = part_refine.s[indx_sorted]
             particles.delta[mask_part] = part_refine.delta[indx_sorted]
-            particles.psigma[mask_part] = part_refine.psigma[indx_sorted]
+            particles.ptau[mask_part] = part_refine.ptau[indx_sorted]
             particles.rvv[mask_part] = part_refine.rvv[indx_sorted]
             particles.rpp[mask_part] = part_refine.rpp[indx_sorted]
             particles.p0c[mask_part] = part_refine.p0c[indx_sorted]
@@ -336,7 +372,8 @@ def build_interp_tracker(_buffer, s0, s1, s_interp, aper_0, aper_1, aper_interp,
             line=Line(elements=ele_all),
             track_kernel=_trk_gen.track_kernel,
             element_classes=_trk_gen.element_classes,
-            reset_s_at_end_turn=0)
+            reset_s_at_end_turn=0,
+            global_xy_limit=_trk_gen.global_xy_limit)
 
     return interp_tracker
 
