@@ -1,3 +1,9 @@
+# copyright ############################### #
+# This file is part of the Xtrack Package.  #
+# Copyright (c) CERN, 2021.                 #
+# ######################################### #
+
+import logging
 import numpy as np
 
 import xobjects as xo
@@ -6,7 +12,7 @@ import xpart as xp
 from scipy.optimize import fsolve
 from scipy.constants import c as clight
 
-from .linear_normal_form import compute_linear_normal_form
+from . import linear_normal_form as lnf
 
 import xtrack as xt # To avoid circular imports
 
@@ -16,11 +22,22 @@ DEFAULT_STEPS_R_MATRIX = {
     'dzeta':1e-6, 'ddelta':1e-7
 }
 
+log = logging.getLogger(__name__)
+
+class ClosedOrbitSearchError(Exception):
+    pass
+
 def find_closed_orbit(tracker, particle_co_guess=None, particle_ref=None,
-                      co_search_settings=None):
+                      co_search_settings=None, delta_zeta=0):
 
     if particle_co_guess is None:
-        assert particle_ref is not None
+        if particle_ref is None:
+            if tracker.particle_ref is not None:
+                particle_ref = tracker.particle_ref
+            else:
+                raise ValueError(
+                    "Either `particle_co_guess` or `particle_ref` must be provided")
+
         particle_co_guess = particle_ref.copy()
         particle_co_guess.x = 0
         particle_co_guess.px = 0
@@ -32,7 +49,6 @@ def find_closed_orbit(tracker, particle_co_guess=None, particle_ref=None,
         particle_co_guess.at_element = 0
         particle_co_guess.at_turn = 0
     else:
-        assert particle_ref is None
         particle_ref = particle_co_guess
 
     if co_search_settings is None:
@@ -45,18 +61,26 @@ def find_closed_orbit(tracker, particle_co_guess=None, particle_ref=None,
     particle_co_guess = particle_co_guess.copy(
                         _context=tracker._buffer.context)
 
-    (res, infodict, ier, mesg
-        ) = fsolve(lambda p: p - _one_turn_map(p, particle_co_guess, tracker),
-              x0=np.array([particle_co_guess._xobject.x[0],
-                           particle_co_guess._xobject.px[0],
-                           particle_co_guess._xobject.y[0],
-                           particle_co_guess._xobject.py[0],
-                           particle_co_guess._xobject.zeta[0],
-                           particle_co_guess._xobject.delta[0]]),
-              full_output=True,
-              **co_search_settings)
-    fsolve_info = {
-        'res': res, 'info': infodict, 'ier': ier, 'mesg': mesg}
+    for shift_factor in [0, 1.]: # if not found at first attempt we shift slightly the starting point
+        if shift_factor>0:
+            log.warning('Need second attempt on closed orbit search')
+        (res, infodict, ier, mesg
+            ) = fsolve(lambda p: p - _one_turn_map(p, particle_co_guess, tracker, delta_zeta),
+                x0=np.array([particle_co_guess._xobject.x[0] + shift_factor * 1e-5,
+                            particle_co_guess._xobject.px[0] + shift_factor * 1e-7,
+                            particle_co_guess._xobject.y[0] + shift_factor * 1e-5,
+                            particle_co_guess._xobject.py[0] + shift_factor * 1e-7,
+                            particle_co_guess._xobject.zeta[0] + shift_factor * 1e-4,
+                            particle_co_guess._xobject.delta[0] + shift_factor * 1e-5]),
+                full_output=True,
+                **co_search_settings)
+        fsolve_info = {
+            'res': res, 'info': infodict, 'ier': ier, 'mesg': mesg}
+        if ier == 1:
+            break
+
+    if ier != 1:
+        raise ClosedOrbitSearchError
 
     particle_on_co = particle_co_guess.copy()
     particle_on_co.x = res[0]
@@ -107,6 +131,8 @@ def compute_one_turn_matrix_finite_differences(
             zeta =  [0.,  0., 0.,  0., dzeta,     0.,  0.,   0.,  0.,   0., -dzeta,      0.],
             delta = [0.,  0., 0.,  0.,    0., ddelta,  0.,   0.,  0.,   0.,     0., -ddelta],
             )
+    dpzeta = float(context.nparray_from_context_array(
+        (part_temp.ptau[5] - part_temp.ptau[11])/2/part_temp.beta0[0]))
     if particle_on_co._xobject.at_element[0]>0:
         part_temp.s[:] = particle_on_co._xobject.s[0]
         part_temp.at_element[:] = particle_on_co._xobject.at_element[0]
@@ -125,28 +151,75 @@ def compute_one_turn_matrix_finite_differences(
     temp_mat[2, :] = context.nparray_from_context_array(part_temp.y)
     temp_mat[3, :] = context.nparray_from_context_array(part_temp.py)
     temp_mat[4, :] = context.nparray_from_context_array(part_temp.zeta)
-    temp_mat[5, :] = context.nparray_from_context_array(part_temp.delta)
+    temp_mat[5, :] = context.nparray_from_context_array(
+                                part_temp.ptau/part_temp.beta0) # pzeta
 
     RR = np.zeros(shape=(6, 6), dtype=np.float64)
 
-    for jj, dd in enumerate([dx, dpx, dy, dpy, dzeta, ddelta]):
+    for jj, dd in enumerate([dx, dpx, dy, dpy, dzeta, dpzeta]):
         RR[:, jj] = (temp_mat[:, jj] - temp_mat[:, jj+6])/(2*dd)
 
     return RR
 
-def _build_auxiliary_tracker_with_extra_markers(tracker, at_s, marker_prefix):
+def _behaves_like_drift(ee):
+    return (hasattr(ee, 'behaves_like_drift') and ee.behaves_like_drift)
+
+
+def _build_auxiliary_tracker_with_extra_markers(tracker, at_s, marker_prefix,
+                                                algorithm='auto'):
+
+    assert algorithm in ['auto', 'insert', 'regen_all_drift']
+    if algorithm == 'auto':
+        if len(at_s)<10:
+            algorithm = 'insert'
+        else:
+            algorithm = 'regen_all_drifts'
 
     auxline = xt.Line(elements=list(tracker.line.elements).copy(),
                       element_names=list(tracker.line.element_names).copy())
 
     names_inserted_markers = []
+    markers = []
     for ii, ss in enumerate(at_s):
         nn = marker_prefix + f'{ii}'
-        auxline.insert_element(element=xt.Drift(length=0),
-                            name=nn,
-                            at_s=ss
-                            )
         names_inserted_markers.append(nn)
+        markers.append(xt.Drift(length=0))
+
+    if algorithm == 'insert':
+        for nn, mm, ss in zip(names_inserted_markers, markers, at_s):
+            auxline.insert_element(element=mm, name=nn, at_s=ss)
+    elif algorithm == 'regen_all_drifts':
+        s_elems = auxline.get_s_elements()
+        s_keep = []
+        enames_keep = []
+        for ss, nn in zip(s_elems, auxline.element_names):
+            if not (_behaves_like_drift(auxline[nn]) and np.abs(auxline[nn].length)>0):
+                s_keep.append(ss)
+                enames_keep.append(nn)
+                assert not xt.line._is_thick(auxline[nn]) or auxline[nn].length == 0
+
+        s_keep.extend(list(at_s))
+        enames_keep.extend(names_inserted_markers)
+
+        ind_sorted = np.argsort(s_keep)
+        s_keep = np.take(s_keep, ind_sorted)
+        enames_keep = np.take(enames_keep, ind_sorted)
+
+        i_new_drift = 0
+        new_enames = []
+        new_ele_dict = auxline.element_dict.copy()
+        new_ele_dict.update({nn: ee for nn, ee in zip(names_inserted_markers, markers)})
+        s_curr = 0
+        for ss, nn in zip(s_keep, enames_keep):
+            if ss > s_curr + 1e-6:
+                new_drift = xt.Drift(length=ss-s_curr)
+                new_dname = f'_auxrift_{i_new_drift}'
+                new_ele_dict[new_dname] = new_drift
+                new_enames.append(new_dname)
+                i_new_drift += 1
+                s_curr = ss
+            new_enames.append(nn)
+        auxline = xt.Line(elements=new_ele_dict, element_names=new_enames)
 
     auxtracker = xt.Tracker(
         _buffer=tracker._buffer,
@@ -169,6 +242,8 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
         particle_co_guess=None, steps_r_matrix=None,
         co_search_settings=None, at_elements=None, at_s=None,
         eneloss_and_damping=False,
+        matrix_responsiveness_tol=lnf.DEFAULT_MATRIX_RESPONSIVENESS_TOL,
+        matrix_stability_tol=lnf.DEFAULT_MATRIX_STABILITY_TOL,
         symplectify=False):
 
     if at_s is not None:
@@ -201,9 +276,11 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
 
     context = tracker._buffer.context
 
-    part_on_co = tracker.find_closed_orbit(particle_co_guess=particle_co_guess,
-                                        particle_ref=particle_ref,
-                                        co_search_settings=co_search_settings)
+    part_on_co = tracker.find_closed_orbit(
+                                particle_co_guess=particle_co_guess,
+                                particle_ref=particle_ref,
+                                co_search_settings=co_search_settings)
+
     RR = tracker.compute_one_turn_matrix_finite_differences(
                                                 steps_r_matrix=steps_r_matrix,
                                                 particle_on_co=part_on_co)
@@ -211,7 +288,9 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
     gemitt_x = nemitt_x/part_on_co._xobject.beta0[0]/part_on_co._xobject.gamma0[0]
     gemitt_y = nemitt_y/part_on_co._xobject.beta0[0]/part_on_co._xobject.gamma0[0]
 
-    W, Winv, Rot = compute_linear_normal_form(RR, symplectify=symplectify)
+    W, Winv, Rot = lnf.compute_linear_normal_form(RR, symplectify=symplectify,
+                                responsiveness_tol=matrix_responsiveness_tol,
+                                stability_tol=matrix_stability_tol)
 
     s = np.array(tracker.line.get_s_elements())
 
@@ -235,6 +314,8 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
         particle_on_co=part_on_co,
         scale_with_transverse_norm_emitt=(nemitt_x, nemitt_y),
         R_matrix=RR,
+        matrix_responsiveness_tol=matrix_responsiveness_tol,
+        matrix_stability_tol=matrix_stability_tol,
         symplectify=symplectify)
 
     part_for_twiss = xp.Particles.merge([part_for_twiss, part_disp])
@@ -247,22 +328,24 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
     py_co = tracker.record_last_track.py[4, :].copy()
     zeta_co = tracker.record_last_track.zeta[4, :].copy()
     delta_co = tracker.record_last_track.delta[4, :].copy()
-    psigma_co = tracker.record_last_track.psigma[4, :].copy()
+    ptau_co = tracker.record_last_track.ptau[4, :].copy()
 
     x_disp_minus = tracker.record_last_track.x[5, :].copy()
     y_disp_minus = tracker.record_last_track.y[5, :].copy()
     px_disp_minus = tracker.record_last_track.px[5, :].copy()
     py_disp_minus = tracker.record_last_track.py[5, :].copy()
+    delta_disp_minus = tracker.record_last_track.delta[5, :].copy()
 
     x_disp_plus = tracker.record_last_track.x[6, :].copy()
     y_disp_plus = tracker.record_last_track.y[6, :].copy()
     px_disp_plus = tracker.record_last_track.px[6, :].copy()
     py_disp_plus = tracker.record_last_track.py[6, :].copy()
+    delta_disp_plus = tracker.record_last_track.delta[6, :].copy()
 
-    dx = (x_disp_plus-x_disp_minus)/delta_disp/2
-    dy = (y_disp_plus-y_disp_minus)/delta_disp/2
-    dpx = (px_disp_plus-px_disp_minus)/delta_disp/2
-    dpy = (py_disp_plus-py_disp_minus)/delta_disp/2
+    dx = (x_disp_plus-x_disp_minus)/(delta_disp_plus - delta_disp_minus)
+    dy = (y_disp_plus-y_disp_minus)/(delta_disp_plus - delta_disp_minus)
+    dpx = (px_disp_plus-px_disp_minus)/(delta_disp_plus - delta_disp_minus)
+    dpy = (py_disp_plus-py_disp_minus)/(delta_disp_plus - delta_disp_minus)
 
     W4 = np.zeros(shape=(4,4,len(s)), dtype=np.float64)
     W4[0, :, :] = (tracker.record_last_track.x[:4, :] - x_co) / scale_transverse_x
@@ -294,13 +377,18 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
                 zeta=part_on_co._xobject.zeta[0], delta=delta_chrom,
                 particle_on_co=part_on_co,
                 scale_with_transverse_norm_emitt=(nemitt_x, nemitt_y),
-                R_matrix=RR, symplectify=symplectify)
+                R_matrix=RR,
+                matrix_stability_tol=matrix_stability_tol,
+                matrix_responsiveness_tol=matrix_responsiveness_tol,
+                symplectify=symplectify)
     RR_chrom_plus = tracker.compute_one_turn_matrix_finite_differences(
                                             particle_on_co=part_chrom_plus.copy(),
                                             steps_r_matrix=steps_r_matrix)
     (WW_chrom_plus, WWinv_chrom_plus, Rot_chrom_plus
-        ) = compute_linear_normal_form(RR_chrom_plus,
-                                          symplectify=symplectify)
+        ) = lnf.compute_linear_normal_form(RR_chrom_plus,
+                                        responsiveness_tol=matrix_responsiveness_tol,
+                                        stability_tol=matrix_stability_tol,
+                                        symplectify=symplectify)
     qx_chrom_plus = np.angle(np.linalg.eig(Rot_chrom_plus)[0][0])/(2*np.pi)
     qy_chrom_plus = np.angle(np.linalg.eig(Rot_chrom_plus)[0][2])/(2*np.pi)
 
@@ -310,13 +398,18 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
                 zeta=part_on_co._xobject.zeta[0], delta=-delta_chrom,
                 particle_on_co=part_on_co,
                 scale_with_transverse_norm_emitt=(nemitt_x, nemitt_y),
-                R_matrix=RR, symplectify=symplectify)
+                R_matrix=RR,
+                matrix_responsiveness_tol=matrix_responsiveness_tol,
+                matrix_stability_tol=matrix_stability_tol,
+                symplectify=symplectify)
     RR_chrom_minus = tracker.compute_one_turn_matrix_finite_differences(
                                         particle_on_co=part_chrom_minus.copy(),
                                         steps_r_matrix=steps_r_matrix)
     (WW_chrom_minus, WWinv_chrom_minus, Rot_chrom_minus
-        ) = compute_linear_normal_form(RR_chrom_minus,
-                                          symplectify=symplectify)
+        ) = lnf.compute_linear_normal_form(RR_chrom_minus,
+                                          symplectify=symplectify,
+                                          stability_tol=matrix_stability_tol,
+                                          responsiveness_tol=matrix_responsiveness_tol)
     qx_chrom_minus = np.angle(np.linalg.eig(Rot_chrom_minus)[0][0])/(2*np.pi)
     qy_chrom_minus = np.angle(np.linalg.eig(Rot_chrom_minus)[0][2])/(2*np.pi)
 
@@ -346,9 +439,8 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
     T_rev = circumference/clight/beta0
 
     if eneloss_and_damping:
-        diff_psigma = np.diff(psigma_co)
-        energy0 = part_on_co.mass0 * part_on_co._xobject.gamma0[0]
-        eloss_turn = -(sum(diff_psigma[diff_psigma<0]) * energy0)
+        diff_ptau = np.diff(ptau_co)
+        eloss_turn = -sum(diff_ptau[diff_ptau<0]) * part_on_co._xobject.p0c[0]
 
         # Get eigenvalues
         w0, v0 = np.linalg.eig(RR)
@@ -359,6 +451,7 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
         eigenvals = np.array([w0[ii*2] for ii in indx])
 
         # Damping constants and partition numbers
+        energy0 = part_on_co.mass0 * part_on_co._xobject.gamma0[0]
         damping_constants_turns = -np.log(np.abs(eigenvals))
         damping_constants_s = damping_constants_turns / T_rev
         partition_numbers = (
@@ -380,7 +473,7 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
         'py': py_co,
         'zeta': zeta_co,
         'delta': delta_co,
-        'psigma': psigma_co,
+        'ptau': ptau_co,
         'betx': betx,
         'bety': bety,
         'alfx': alfx,
@@ -406,6 +499,7 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
         'R_matrix': RR,
         'particle_on_co':part_on_co.copy(_context=xo.context_default)
     }
+    twiss_res['particle_on_co']._fsolve_info = part_on_co._fsolve_info
 
     if eneloss_and_damping:
         twiss_res.update(eneloss_damp_res)
@@ -432,13 +526,13 @@ def twiss_from_tracker(tracker, particle_ref, r_sigma=0.01,
 
     return twiss_res
 
-def _one_turn_map(p, particle_ref, tracker):
+def _one_turn_map(p, particle_ref, tracker, delta_zeta):
     part = particle_ref.copy()
     part.x = p[0]
     part.px = p[1]
     part.y = p[2]
     part.py = p[3]
-    part.zeta = p[4]
+    part.zeta = p[4] + delta_zeta
     part.delta = p[5]
 
     tracker.track(part)
